import numpy as np
import pandas as pd
import pandas_ta as ta
from technical import qtpylib
from freqtrade.strategy import IStrategy, stoploss_from_open
from freqtrade.persistence import Trade
from datetime import datetime



class VWAPStrategy(IStrategy):
        
    custom_stoploss_per_trade = {}
    recent_trades = set()

    INTERFACE_VERSION = 2

    timeframe = '5m'

    minimal_roi = {
    "0": 1
    }

    stoploss = -0.2


    trailing_stop = False
    trailing_stop_positive = 0.01
    trailing_stop_positive_offset = 0.03
    trailing_only_offset_is_reached = False
        

    def custom_stoploss(self, pair: str, trade: 'Trade', current_time: datetime, current_rate: float, current_profit: float, after_fill: bool, **kwargs) -> float:

        # Check if a custom stoploss has already been set
        if trade.id in self.custom_stoploss_per_trade:
            # If profit is greater than 2%, update the stoploss to half of the current profit
            if current_profit >= 0.02:
                updated_stoploss = current_profit / 2
                # Update only if the new stoploss is tighter than the existing one
                if updated_stoploss < self.custom_stoploss_per_trade[trade.id]:
                    self.custom_stoploss_per_trade[trade.id] = updated_stoploss
                    return updated_stoploss

            # If profit is greater than 1%, and not already handled by the above block, consider updating the stoploss
            elif current_profit >= 0.01:
                updated_stoploss = current_profit / 2
                # Update only if the new stoploss is tighter than the existing one
                if updated_stoploss < self.custom_stoploss_per_trade[trade.id]:
                    self.custom_stoploss_per_trade[trade.id] = updated_stoploss
                    return updated_stoploss
            
            # Return the already set custom stoploss if no updates were made
            return self.custom_stoploss_per_trade[trade.id]

        # Retrieve the analyzed dataframe for the current currency pair and a given timeframe
        dataframe, _ = self.dp.get_analyzed_dataframe(pair, self.timeframe)
        last_candle = dataframe.iloc[-1].squeeze()

        # Set the ATR stoploss when the trade opens
        if pd.notna(last_candle['buy_atr']):
            # Calculate stoploss based on 'buy_atr' if it's not NaN
            ratio = last_candle['buy_atr'] / trade.open_rate
            stoploss_from_open = trade.open_rate * (1 - ratio)
            self.custom_stoploss_per_trade[trade.id] = stoploss_from_open
            return stoploss_from_open

        # Default stoploss if none of the above conditions are met
        return self.stoploss


    def populate_indicators(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        dataframe.set_index(pd.DatetimeIndex(dataframe["date"]), inplace=True)

        # Calclate ATR
        dataframe['ATR'] = ta.atr(dataframe['high'], dataframe['low'], dataframe['close'], length=150)

        dataframe['ATR-close'] = dataframe['close'] - (dataframe['ATR'] * 3)

        # Calculate and add the Relative Strength Index (RSI) to the dataframe
        dataframe['rsi'] = ta.rsi(dataframe['close'], length=16)

        # Calculate and add the Volume Weighted Average Price (VWAP) to the dataframe
        dataframe['VWAP'] = ta.vwap(dataframe['high'], dataframe['low'], dataframe['close'], dataframe['volume'], anchor='D', offset=None)

        # Calculate the 14-period Simple Moving Average (SMA)
        sma = dataframe['close'].rolling(window=14).mean()

        # Calculate the Standard Deviation for the same period
        std_dev = dataframe['close'].rolling(window=14).std()

        # Calculate the Upper Bollinger Band
        dataframe['upper_band'] = sma + (std_dev * 2.0)

        # Calculate the Lower Bollinger Band
        dataframe['lower_band'] = sma - (std_dev * 2.0)

        # Initialize a list to store VWAP signals and set the number of backcandles to consider
        VWAP_signal = [0] * len(dataframe)
        backcandles = 15

        # Iterate over rows in the dataframe to determine VWAP signals
        for row in range(backcandles, len(dataframe)):
            # Initialize flags for up_trend and down_trend
            up_trend = 1
            down_trend = 1

            # Iterate over a range of previous candles to check the relationship between prices and VWAP
            for i in range(row - backcandles, row + 1):
                # Check if the maximum price (open or close) is greater than or equal to VWAP
                if max(dataframe['open'][i], dataframe['close'][i]) >= dataframe['VWAP'][i]:
                    down_trend = 0 # Set down_trend to 0

                # Check if the minimum price (open or close) is less than or equal to VWAP
                if min(dataframe['open'][i], dataframe['close'][i]) <= dataframe['VWAP'][i]:
                    up_trend = 0 # Set up_trend to 0

            # Based on up_trend and down_trend, assign a VWAP signal
            if up_trend == 1 and down_trend == 1:
                VWAP_signal[row] = 3 # Neutral signal
            elif up_trend == 1:
                VWAP_signal[row] = 2 # Upward trend signal - 15 candles above the VWAP line
            elif down_trend == 1:
                VWAP_signal[row] = 1 # Downward trend signal - 15 candles below the VWAP line

        # Add the calculated VWAP signals to the dataframe
        dataframe['VWAP_signal'] = VWAP_signal

        return dataframe

    
    def populate_entry_trend(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        condition = (
            (dataframe['volume'] > 0) & # Buy when volume > 0
            (dataframe['VWAP_signal'] == 2) & # Buy when VWAP_signal is 2 (15 candles above VWAP line - indicating bullish trend)
            (dataframe['rsi'] < 45) & # Buy when rsi < 45
            (dataframe['close'] <= dataframe['lower_band']) & # Buy when the current closing price is less than or equal to the current lower bband
            (dataframe['close'].shift(1) <= dataframe['lower_band'].shift(1)) & # Buy when the previous closing price was less than or equal to the previous lower bband
            (dataframe['lower_band'] != dataframe['upper_band']) # Make sure lower and upper bband is not the same (no momentum)
        )

        # If all conditions are True for a given row, the 'buy' column for that row is set to 1,
        # and the 'buy_atr' column is set to the corresponding 'ATR' value.
        dataframe.loc[condition, 'buy'] = 1
        dataframe.loc[condition, 'buy_atr'] = dataframe.loc[condition, 'ATR-close']

        return dataframe


    def populate_exit_trend(self, dataframe: pd.DataFrame, metadata: dict) -> pd.DataFrame:

        dataframe.loc[
            (
                (dataframe['volume'] > 0) & # Sell when volume > 0
                (dataframe['VWAP_signal'] == 2) & # Sell when VWAP_signal is 2 (15 candles above VWAP line - indicating bullish trend)
                (dataframe['close'] >= dataframe['upper_band']) & # Sell when closing price is the same or more than upper bband'
                (dataframe['rsi'] > 55) & # Sell when rsi > 55
                (dataframe['rsi'] <= 90) & # Sell when rsi <= 90
                (dataframe['lower_band'] != dataframe['upper_band']) # Make sure lower and upper bband is not the same (no momentum)
            ),
            # If all conditions are True for a given row, the 'sell' column for that row is set to 1
            'sell'
        ] = 1

        return dataframe
